"""
Copyright (c) 2010, Cambridge Silicon Radio Ltd.
Written by Emilio Monti <emilmont@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.
3. Neither the name of the Cambridge Silicon Radio Ltd. nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from array import array
from MozSymbolsUploader.elf.stream import ParseError
from MozSymbolsUploader.elf.enums import SHT, SHF, MACHINE, ELFCLASS, ELFDATA
import os

class Header(object):
    
    def __init__(self, elf):
        elf.constant(4, "\x7fELF")
        self.elfclass = elf.u08()
        elf.set_bits(self.elfclass)
        
        self.elfdata = elf.u08()
        elf.set_endianness(self.elfdata)
        
        self.version = elf.u08()
        elf.skip(9)
        
        self.type = elf.u16()
        self.machine = elf.u16()
        self.version = elf.u32()
        if self.elfclass == ELFCLASS.ELFCLASS32:
            self.entry = elf.u32()
            self.ph_offset = elf.u32()
            self.sh_offset = elf.u32()
        else:
            self.entry = elf.u64()
            self.ph_offset = elf.u64()
            self.sh_offset = elf.u64()
        self.flags = elf.u32()
        self.header_size = elf.u16()
        self.ph_entry_size = elf.u16()
        self.ph_count = elf.u16()
        self.sh_entry_size = elf.u16()
        self.sh_count = elf.u16()
        self.shstrndx = elf.u16()
    
    
    

class SectionHeader(object):
    def __init__(self, elf, index):
        self.elf = elf
        self.index = index
        
        self.name_index = elf.u32()
        self.type = elf.u32()
        if self.elf.header.elfclass == ELFCLASS.ELFCLASS32:
            self.flags = elf.u32()
            self.addr = elf.u32()
            self.offset = elf.u32()
            self.size = elf.u32()
        else:
            self.flags = elf.u64()
            self.addr = elf.u64()
            self.offset = elf.u64()
            self.size = elf.u64()
        self.link = elf.u32()
        self.info = elf.u32()
        if self.elf.header.elfclass == ELFCLASS.ELFCLASS32:
            self.addralign = elf.u32()
            self.entsize = elf.u32()
        else:
            self.addralign = elf.u64()
            self.entsize = elf.u64()
        self._name = None
        self._data = None
        
    def is_loadable(self):
        return self.type == SHT.PROGBITS and self.flags & SHF.ALLOC == SHF.ALLOC
    
    def is_execinstr(self):
        return self.flags & SHF.EXECINSTR == SHF.EXECINSTR
     
    # name property ############################################
    @property
    def name(self):
        if self._name == None:
            self._name = self.elf.shstrtab[self.name_index]
        return self._name
        
    @name.setter
    def name(self,value):
        pass

    # symbols property ############################################
    @property
    def symbols(self):
        return [sym for sym in self.elf.symbols if sym.shndx == self.index]
    
    # data property ############################################
    @property 
    def data(self):
        if self._data == None:
            curr_offset = self.elf.io.tell()
            self.elf.io.seek(self.offset, os.SEEK_SET)
            self._data = self.elf.io.read(self.size)
            self.elf.io.seek(curr_offset, os.SEEK_SET)
        return self._data
    
    @data.setter
    def data(self, data):
        if len(data) != len(self.__data) :
            raise ParseError('Size of new data (%d) mismatch with current '
                                'size (%d)' % (len(data), self.size))
        self._data = data
        curr_offset = self.elf.io.tell()
        self.elf.io.seek(self.offset, os.SEEK_SET)
        self.elf.io.read(self.__data)
        self.elf.io.seek(curr_offset, os.SEEK_SET)

class ProgramHeader(object):
    def __init__(self, elf, index):
        self.elf = elf
        self.index = index
        
        self.type = elf.u32()
        if self.elf.header.elfclass == ELFCLASS.ELFCLASS32:
            self.offset = elf.u32()
            self.vaddr = elf.u32()
            self.paddr = elf.u32()
            self.filesz = elf.u32()
            self.memsz = elf.u32()
            self.flags = elf.u32()
            self.align = elf.u32()
        else:
            self.flags = elf.u32()
            self.offset = elf.u64()
            self.vaddr = elf.u64()
            self.paddr = elf.u64()
            self.filesz = elf.u64()
            self.memsz = elf.u64()
            self.align = elf.u64()

class Symbol(object):
    LENGTH = 16
    
    def __init__(self, elf, index):
        self.elf = elf
        self.index = index
        
        self.name_index = elf.u32()
        self.value = elf.u32()
        self.size = elf.u32()
        self.info = elf.u08()
        self.other = elf.u08()
        self.shndx = elf.u16()
        
        self._name = None
        
    # name property ############################################
    @property
    def name(self):
        if self._name == None:
            self._name = self.elf.shstrtab[self.name_index]
        return self._name
        
    @name.setter
    def name(self,value):
        pass
    
    def get_bind(self):
        return self.info >> 4
    
    def get_type(self):
        return self.info & 0xF


class StringTable(object):
    def __init__(self, stream, offset, size):
        self.offset = offset
        stream.seek(offset)
        self.table = array('c', stream.read(size))
        self.max = len(self.table)
    
    def __getitem__(self, key):
        if (key >= self.max):
            raise ParseError('The required index is out of the table: (0x%x) '
                        '+%d (max=%d)' % (self.offset, key, self.max))
        i = self.table[key:].index('\x00') + key
        return self.table[key:i].tostring()
